#!   /usr/bin/env   python
#    coding: utf8

# Copyright CERN, 2011
# Author: Matthieu Cattin <matthieu.cattin@cern.ch>
# Licence: GPL v2 or later.
# Website: http://www.ohwr.org

import sys
import rr
import time
import os
from numpy import *
from pylab import *
from ctypes import *

from ptsexcept import *

import spec_fmc_adc
import fmc_adc
import calibr_box
import find_usb_tty
import cp210x_eeprom
from PAGE.Agilent33250A import *
from PAGE.SineWaveform import *


"""
test19: Calibration

Note: Requires test00.py to run first to load the firmware!
"""

# Calibration box vendor and product IDs
BOX_USB_VENDOR_ID = 0x10c4 # Cygnal Integrated Products, Inc.
BOX_USB_PRODUCT_ID = 0xea60 # CP210x Composite Device
# Agilent AWG serial access vendor and product IDs
AWG_USB_VENDOR_ID = 0x0403 # Future Technology Devices International, Ltd
AWG_USB_PRODUCT_ID = 0x6001 # FT232 USB-Serial (UART) IC

AWG_BAUD = 57600

NB_CHANNELS = 4

AWG_SET_SLEEP = 0.1
SSR_SET_SLEEP = 0.005
BOX_SET_SLEEP = 0.01
DAC_SET_SLEEP = 0.01

ACQ_TIMEOUT = 10

MAX_FIRMWARE_RELOAD = 10

PRE_TRIG_SAMPLES = 100
POST_TRIG_SAMPLES = 100000
NB_SHOTS = 1

ACQ_LENGTH = 50000 # in samples

ADC_NBITS = 16 # ADC chip is 14 bits, but shifted to 16 bits in the firmware
DAC_NBITS = 16
DAC_FS = 10 # DAC full scale range is 10V

# Full scale volatge values for input ranges
ADC_FS = {'10V':10.0, '1V':1.0, '100mV':0.1}

RANGES = ['10V', '1V', '100mV']

# Number of repeatition of the measurement, then the results are averaged
REPEAT = 2

ADC_LSB = {'10V':10.0/2**14, '1V':1.0/2**14, '100mV':0.1/2**14}


def load_firmware(default_directory):
    print('Load firmware to FPGA')
    path_fpga_loader = '../../../gnurabbit/user/fpga_loader';
    path_firmware = '../firmwares/spec_fmcadc100m14b4cha_test.bin';
    firmware_loader = os.path.join(default_directory, path_fpga_loader)
    bitstream = os.path.join(default_directory, path_firmware)
    print firmware_loader + ' ' + bitstream + '\n'
    os.system( firmware_loader + ' ' + bitstream )
    time.sleep(2);

def disconnect_channels(fmc):
    for i in range(1,NB_CHANNELS+1):
        fmc.set_ssr(i, 0x00)
        time.sleep(SSR_SET_SLEEP)

def fmc_adc_init(spec, fmc):
    print('Initialise FMC board\n')
    # Reset offset DACs
    fmc.dc_offset_reset()
    # Make sure all switches are OFF
    disconnect_channels(fmc)
    # Reset offset DACs
    fmc.dc_offset_reset()
    # Set trigger
    fmc.set_soft_trig()
    # Set acquisition
    fmc.set_pre_trig_samples(PRE_TRIG_SAMPLES)
    fmc.set_post_trig_samples(POST_TRIG_SAMPLES)
    fmc.set_shots(NB_SHOTS)
    # Print configuration
    #fmc.print_adc_core_config()

# Converts two's complement hex to signed
def hex2signed(value):
    if(value & 0x8000):
        return -((~value & 0xFFFF) + 1)
    else:
        return value

# Converts signed to two's complement hex
def signed2hex(value):
    if value < 0:
        return (((abs(value) ^ 0xffff) + 1) & 0xffff)
    else:
        return value

# Converts digital value to volts
def digital2volt(value, full_scale, nb_bit):
    return float(value) * float(full_scale)/2**nb_bit

# Converts volts to digital value with half full range offset
def volt2digital(value, full_scale, nb_bit):
    digital = (value + full_scale/2) * 2**nb_bit/full_scale
    if(digital > 2**nb_bit - 1):
        digital = 2**nb_bit - 1
    if(digital < 0):
        digital = 0
    #print('volt2digital: %2.9f > %2.9f')%(value,digital)
    return int(digital)

# Converts volts to digital value
def volt2digital_without_offset(value, full_scale, nb_bit):
    if(value > (2**nb_bit)/2 - 1):
        value = (2**nb_bit)/2 - 1
    if(value < -((2**nb_bit)/2)):
        value = -((2**nb_bit)/2)
    digital = (value) * 2**nb_bit/full_scale
    #print('volt2digital: %2.9f > %2.9f')%(value,digital)
    return int(digital)

def acq_channel(spec_fmc, fmc, ch, adc_fs, adc_nbits=16, pause=0.01):
    # Make sure no acquisition is running
    fmc.stop_acq()
    # Start acquisition
    fmc.start_acq()
    #time.sleep(pause)
    # Trigger
    fmc.sw_trig()
    # Wait end of acquisition
    timeout = 0
    time.sleep(0.001)
    while('IDLE' != fmc.get_acq_fsm_state()):
        time.sleep(0.01)
        timeout += 1
        if(ACQ_TIMEOUT < timeout):
            print "Acquisition timeout. Missing trigger?."
            print "Acq FSM state: %s"%fmc.get_acq_fsm_state()
            return 1
    # Retrieve data trough DMA
    trig_pos = fmc.get_trig_pos()
    # Enable "DMA done" iinterrupt
    spec_fmc.set_irq_en_mask(0x1)
    # Read ACQ_LENGTH samples after the trigger for all channels
    channels_data = spec_fmc.get_data((trig_pos<<3), ACQ_LENGTH*8)
    # Disable "DMA done" iinterrupt
    spec_fmc.set_irq_en_mask(0x0)
    channels_data = [hex2signed(item) for item in channels_data]
    channels_data = [digital2volt(item,adc_fs,adc_nbits) for item in channels_data]
    channel_data = channels_data[ch-1::4]
    return mean(channel_data)

def plot_all(data, mean, ylimit):
    sample = arange(len(data)/4)
    clf()
    plot(sample, data[0::4], 'b', label='Channel 1')
    plot(sample, data[1::4], 'g', label='Channel 2')
    plot(sample, data[2::4], 'c', label='Channel 3')
    plot(sample, data[3::4], 'm', label='Channel 4')
    plot(sample, [mean[0]]*len(sample), 'r')
    plot(sample, [mean[1]]*len(sample), 'r')
    plot(sample, [mean[2]]*len(sample), 'r')
    plot(sample, [mean[3]]*len(sample), 'r')
    ylim(-ylimit-(ylimit/10.0), ylimit+(ylimit/10.0))
    grid(which='both')
    legend()
    draw()
    show()
    return 0

def set_offset_dac(fmc, channel, offset_volt, dac_fs=10.0, dac_nbits=16, dac_corr_flag=False):
    dac_v = offset_volt
    dac_d = volt2digital(dac_v,dac_fs,dac_nbits)
    #print('DAC value: 0x%X (%fV)')%(dac_d, dac_v)
    if(True == dac_corr_flag):
        fmc.set_dc_offset_corrected(channel,dac_d)
    else:
        fmc.set_dc_offset(channel,dac_d)
    time.sleep(DAC_SET_SLEEP)

def make_meas(spec_fmc, fmc, box, dac_value, in_value, in_range, adc_fs, repeat):
    ch_meas = []
    ch_diff = []
    for ch in range(1,NB_CHANNELS+1):
        # Set offset DAC
        set_offset_dac(fmc, ch, dac_value)
        # Configure calibration box
        if in_value == 0.0:
            # Configure analogue input, ground input
            fmc.set_input_range(ch, ('CAL_'+in_range))
            time.sleep(SSR_SET_SLEEP)
            # Connect AWG to current channel
            box.select_output_ch(ch)
            time.sleep(BOX_SET_SLEEP)
        else:
            # Configure analogue input range
            fmc.set_input_range(ch, in_range)
            time.sleep(SSR_SET_SLEEP)
            # Connect Vref for selected input range to current channel
            # Note that channel in parameter is the one connected to AWG
            if ch == 1:
                awg_ch = 4
            else:
                awg_ch = ch-1
            box.select_output_ch(awg_ch, in_range)
            time.sleep(BOX_SET_SLEEP)
        # Perform an acquisition
        meas = []
        for i in range(repeat):
            meas.append(acq_channel(spec_fmc, fmc, ch, adc_fs))
        # Calculate mean and delta of acquisitions
        ch_meas.append(mean(meas))
        ch_diff.append(max(meas)-min(meas))
    return ch_meas, ch_diff


# Calculates ADC + input stage gain
def calc_ga(Vm1, Vm2, Vref1):
    return ((Vm2-Vm1)/Vref1)

# Calculates ADC + input stage offset
def calc_oa(Vm2, Vm3, Vm4):
    return (Vm2 + Vm3 - Vm4)

# Calculates DAC gain
def calc_gd(Vm1, Vm2, Vm3, Vref1, Vref2):
    return ((Vref1*(Vm3-Vm1))/(Vref2*(Vm1-Vm2)))

# Calculates DAC offset
def calc_od(Vm1, Vm2, Vm3, Vm4, Vref1):
    return ((Vref1*(Vm1-Vm2-Vm3+Vm4))/(Vm1-Vm2))


def main (default_directory = '.'):

    # Load firmware to FPGA
    load_firmware(default_directory)

    # Objects declaration
    spec = rr.Gennum() # bind to the SPEC board
    spec_fmc = spec_fmc_adc.CSpecFmcAdc100Ms(spec)
    fmc = fmc_adc.CFmcAdc100Ms(spec)
    usb_tty = find_usb_tty.CttyUSB()
    awg_tty = usb_tty.find_usb_tty(AWG_USB_VENDOR_ID, AWG_USB_PRODUCT_ID)
    box_tty = usb_tty.find_usb_tty(BOX_USB_VENDOR_ID, BOX_USB_PRODUCT_ID)
    #print "AWG:%s"%awg_tty[0]
    #print "BOX:%s"%box_tty[0]
    gen = Agilent33250A(device=awg_tty[0], bauds=AWG_BAUD)
    sine = SineWaveform()
    box = calibr_box.CCalibr_box(box_tty[0])
    box_eeprom = cp210x_eeprom.CCP210x_Eeprom("%X"%BOX_USB_VENDOR_ID, "%X"%BOX_USB_PRODUCT_ID)

    # Initialise fmc adc
    fmc_adc_init(spec, fmc)

    # Connect to AWG
    gen.connect()

    # Switch AWG output and sync OFF
    gen.output = False
    gen.sync = False

    # Measure FMC and carrier temperature
    print('SPEC temperature: %3.3f°C') % spec_fmc.get_temp()
    print('FMC temperature : %3.3f°C') % fmc.get_temp()

    # Open all switches, reset offset DAC to mid-scale (0V)
    for channel in range(1,NB_CHANNELS+1):
        fmc.set_input_range(channel, 'OPEN')
        fmc.set_input_term(channel, 'OFF')
    fmc.dc_offset_reset()

    # Get calibration data from calibration box
    box_calibr_data = box_eeprom.get_calibr_data()
    print "\nCalibration box data:"
    print box_calibr_data

    # Offset DAC reference voltage
    offset_dac_data = {'10V':4.096, '1V':0.4096, '100mV':0.04096}
    print "\nOffset DAC reference volatges:"
    print offset_dac_data

    # Dict to store ADC calibration data
    adc_corr_data = {'10V':{'offset':[],'gain':[]},
                     '1V':{'offset':[],'gain':[]},
                     '100mV':{'offset':[],'gain':[]}}

    # Dict to store DAC calibration data
    dac_corr_data = {'10V':{'offset':[],'gain':[]},
                     '1V':{'offset':[],'gain':[]},
                     '100mV':{'offset':[],'gain':[]}}


    ############################################################################
    # Calibration
    ############################################################################

    error = 0
    for IN_RANGE in RANGES:

        print "\n%s range calibration\n----------------------------------\n"%(IN_RANGE)

        print('SPEC temperature: %3.3f°C') % spec_fmc.get_temp()
        print('FMC temperature : %3.3f°C') % fmc.get_temp()

        print "\nADC LSB:%1.9f"%(ADC_LSB[IN_RANGE])

        # Reference voltage for calibration
        vref_in  = float(box_calibr_data[IN_RANGE]) # from box cailbration data in CP2103 EEPROM
        vref_dac = offset_dac_data[IN_RANGE]  # reference voltage to set offset DAC

        # Input and offset DAC volatges for the four measurements
        v_in  = [0.0, vref_in, 0.0,      vref_in]
        v_dac = [0.0, 0.0,     vref_dac, vref_dac]

        #-----------------------------------------------------------------------
        # Make the measurements
        #-----------------------------------------------------------------------
        v_meas = []
        d_meas = []
        error = 0
        for i in range(len(v_in)):
            print "\nMeasurement %d: channel input = %2.8fV, offset DAC = %2.8fV"%((i+1), v_in[i], v_dac[i])
            mean, diff = make_meas(spec_fmc, fmc, box, v_dac[i], v_in[i], IN_RANGE, ADC_FS[IN_RANGE], REPEAT)
            v_meas.append(mean)
            d_meas.append(diff)
            for ch in range(1,NB_CHANNELS+1):
                print "Channel %d: v_meas=%02.9fV delta=%1.9f"%(ch, v_meas[i][ch-1], d_meas[i][ch-1])
                print "   Check voltage stability"
                meas = d_meas[i][ch-1]
                expect = (2*ADC_LSB[IN_RANGE])
                if meas > expect:
                    print "      ERROR: measure:%1.9f expect:%1.9f"%(meas, expect)
                    error += 1
                else:
                    print "      OK"


        #-----------------------------------------------------------------------
        # Calculate gain and offset parameters
        # ga = ADC + input stage gain
        # oa = ADC + input stage offset
        # gd = DAC gain
        # od = DAC offset
        #-----------------------------------------------------------------------
        print "\n%s range correction parameters\n----------------------------------"%(IN_RANGE)

        ga = []
        for ch in range(1,NB_CHANNELS+1):
            ga.append(calc_ga(v_meas[0][ch-1], v_meas[1][ch-1], vref_in))
            print "Channel %d ADC gain coeff: %02.9f"%(ch, ga[ch-1])

        oa = []
        for ch in range(1,NB_CHANNELS+1):
            oa.append(calc_oa(v_meas[1][ch-1], v_meas[2][ch-1], v_meas[3][ch-1]))
            print "Channel %d ADC offset    : %02.9f"%(ch, oa[ch-1])

        gd = []
        for ch in range(1,NB_CHANNELS+1):
            gd.append(calc_gd(v_meas[0][ch-1], v_meas[1][ch-1], v_meas[2][ch-1], vref_in, vref_dac))
            print "Channel %d DAC gain coeff: %02.9f"%(ch, gd[ch-1])

        od = []
        for ch in range(1,NB_CHANNELS+1):
            od.append(calc_od(v_meas[0][ch-1], v_meas[1][ch-1], v_meas[2][ch-1], v_meas[3][ch-1], vref_in))
            print "Channel %d DAC offset    : %02.9f"%(ch, od[ch-1])

        #-----------------------------------------------------------------------
        # Calculate ADC correction register values
        #-----------------------------------------------------------------------
        # 0x8000 corresponds to 1.0
        # Gain correction register is written with 1/gain * 1.0
        adc_corr_data[IN_RANGE]['gain'] = [int(round((1/item)*0x8000)) for item in ga]

        # oa is in volts and has to be converted to digital raw value,
        # to be written in offset correction register
        adc_corr_data[IN_RANGE]['offset'] = [-(volt2digital_without_offset(item,ADC_FS[IN_RANGE],ADC_NBITS)) for item in oa]

        #-----------------------------------------------------------------------
        # Calculate DAC correction register values
        #-----------------------------------------------------------------------
        # 0x8000 corresponds to 1.0
        # Gain correction register is written with 1/gain * 1.0
        dac_corr_data[IN_RANGE]['gain'] = [int(round((1/item)*0x8000)) for item in gd]

        # oa is in volts and has to be converted to digital raw value,
        # to be written in offset correction register
        dac_corr_data[IN_RANGE]['offset'] = [-(volt2digital_without_offset(item,DAC_FS,DAC_NBITS)) for item in od]

    ############################################################################
    # Write ADC correction data tp file
    ############################################################################
    out_filename = "test19_corr_data.txt"
    f_out = open(out_filename, 'w')

    print "\nWrite ADC correction data to %s"%(out_filename)
    for IN_RANGE in RANGES:
        #print "%s range"%(IN_RANGE)
        for item in adc_corr_data[IN_RANGE]['offset']:
            #print '0x%.4X,yes'%(signed2hex(item))
            f_out.write('0x%.4X,yes\n'%(signed2hex(item)))
        for item in adc_corr_data[IN_RANGE]['gain']:
            #print '0x%.4X,yes'%(signed2hex(item))
            f_out.write('0x%.4X,yes\n'%(signed2hex(item)))
    f_out.close()


    ############################################################################
    # Write DAC correction data tp file
    ############################################################################
    out_filename = "test19_dac_corr_data.txt"
    f_out = open(out_filename, 'w')

    print "\nWrite DAC correction data to %s"%(out_filename)
    for IN_RANGE in RANGES:
        #print "%s range"%(IN_RANGE)
        for item in dac_corr_data[IN_RANGE]['offset']:
            #print '0x%.4X,yes'%(signed2hex(item))
            f_out.write('0x%.4X,yes\n'%(signed2hex(item)))
        for item in dac_corr_data[IN_RANGE]['gain']:
            #print '0x%.4X,yes'%(signed2hex(item))
            f_out.write('0x%.4X,yes\n'%(signed2hex(item)))
    f_out.close()


    """
    ############################################################################
    # Format FMC EEPROM data
    ############################################################################
    os.system("python ../../fmceeprom/python/fmc_eeprom_gen.py -i ../../fmceeprom/python/eeprom_input.xml -o test19_formatted_data.txt -r test19_corr_data.txt")

    ############################################################################
    # Write data to FMC EEPROM
    ############################################################################
    in_filename = "test19_formatted_data.txt"
    eeprom_content = []
    eeprom_addr = []
    f_in = open(in_filename,"r+")
    for line in f_in:
        addr,data=line.split()
        eeprom_content.append(int(data,2))
        eeprom_addr.append(int(addr,2))
    fmc.sys_i2c_eeprom_write(eeprom_addr[0],eeprom_content)

    ############################################################################
    # Verify EEPROM data
    ############################################################################
    eeprom_content_read = fmc.sys_i2c_eeprom_read(eeprom_addr[0],len(eeprom_content))
    for i in len(eeprom_content):
        if eeprom_content[i] != eeprom_content_read[i]:
            print "[ERROR] eeprom data corruption @ addr:%d expect:0x%.2X read:0x%.2X"%(i, eeprom_content[i], eeprom_content_read[i])

    """

    # Open all switches, reset offset DAC to mid-scale (0V)
    for channel in range(1,NB_CHANNELS+1):
        fmc.set_input_range(channel, 'OPEN')
        fmc.set_input_term(channel, 'OFF')
    fmc.dc_offset_reset()

    # Close AWG
    gen.close()

    # Check if an error occured during frequency response test
    if(error != 0):
        raise PtsError('An error occured during calibration, check log for details.')


if __name__ == '__main__' :
    main()
